﻿// © ANO PO "KSI", 2016

#include "HashTools.h"

#include <QFileInfo>

#include <filesystem>
#include <fstream>
#include <sstream>

HashTools::HashTools(const std::string& hashType, QObject * parent /*= nullptr*/) 
	: QObject(parent),
	m_hashType(hashType)
{
	wrapperfactory factory;
	m_pHashWrapper = factory.create(hashType);
}

HashTools::~HashTools() {
	delete m_pHashWrapper;
}

void HashTools::createHashList(const QString& dirName, HashFileList* pHashFileList, ICreateHashReport* report)
{
	QDir baseDir(dirName);

	QFileInfoList fileList;

	createFileList(baseDir, &fileList);
	
	pHashFileList->clear();

	if (report) {
		report->start(fileList.size());
	}

	for (const auto& p : fileList) {

		QString relativeFilePath = baseDir.relativeFilePath(QDir::toNativeSeparators(p.filePath()));

		if (report) {
			report->setCurrentFile(relativeFilePath);

			if (report->wasCancelled()) {
				pHashFileList->clear();
				return;
			}
		}

		std::string hash;

		try {
			hash = m_pHashWrapper->getHashFromFile(QDir::toNativeSeparators(p.filePath()).toLocal8Bit().toStdString());
		}
		catch (hlException& e) {
			if (report) {
				pHashFileList->clear();
				report->setError(QString::fromStdString(e.error_message()));
				return;
			}
		}

		if (hash.length() > 0) {
			pHashFileList->push_back(HashFile(
				relativeFilePath.toStdString(),
				hash));
		}
	}

	if (report) {
		report->complete();
	}
}

void HashTools::createFileList(const QDir& dir, QFileInfoList* fileList)
{
	*fileList += dir.entryInfoList(QDir::Files);

	QFileInfoList foundDirs = dir.entryInfoList(QDir::AllDirs | QDir::NoDotAndDotDot);

	for (const auto& p : foundDirs) {
		createFileList(p.filePath(), fileList);
	}
}

void HashTools::updateHashList(HashFileList* pHashFileList)
{
	for (auto& p : *pHashFileList) {
		try {
			p.hash = m_pHashWrapper->getHashFromFile(p.filename);
		}
		catch (const hlException& e) {
			p.hash.clear();
		}
	}
}

std::string HashTools::loadDB(const std::string& dbname, HashFileList* pHashFileList)
{
	std::ifstream dbstream(dbname);
	if (!dbstream) {
		return std::string();
	}

	std::string line;
	while (std::getline(dbstream, line)) {

		std::istringstream iss(line);
		HashFile hf;

		if (!(iss >> hf.hash)) {
			break;
		}

		hf.filename = line.substr(hf.hash.length() + 1);

		pHashFileList->push_back(hf);
	}

	dbstream.close();

	return m_pHashWrapper->getHashFromFile(dbname);
}

std::string HashTools::createDB(const std::string& dbname, const HashFileList& hashFileList)
{
	std::ofstream dbstream(dbname, std::ios::out | std::ios::trunc);
	if (!dbstream) {
		return std::string();
	}

	for (const auto& p : hashFileList) {
		dbstream 
			<< p.hash 
			<< " " 
			<< p.filename
			<< std::endl;
	}

	dbstream.close();

	return m_pHashWrapper->getHashFromFile(dbname);
}

std::string HashTools::hashType() const
{
	return m_hashType;
}

void HashTools::compare(const QDir& baseDir, const HashFileList& newHashFileList, const HashFileList& dbHashFileList, CompareErrList* errors)
{
	using namespace std::tr2;

	errors->clear();

	auto eqfilenames = [&](const std::string& fn1, const std::string& fn2)->bool {
//		too slow :(
// 		QFileInfo fileInfo1, fileInfo2;
// 		fileInfo1.setFile(baseDir, QString::fromLocal8Bit(fn1.c_str()));
// 		fileInfo2.setFile(baseDir, QString::fromLocal8Bit(fn2.c_str()));
// 
// 		return fileInfo1 == fileInfo2;

		return fn1 == fn2;

	};

	for (const auto& p : newHashFileList) {
		bool fileFound = false;
		for (const auto& q : dbHashFileList) {
			if (eqfilenames(p.filename, q.filename))	{

				if (p.hash != q.hash) {
					errors->push_back(std::make_unique<CE_HashDiff>(p.filename, p.hash, q.hash));
				}
				fileFound = true;
				break;
			}
		}
 		if (!fileFound) {
 			errors->push_back(std::make_unique<CE_NewFile>(p.filename));
 		}
	}

	// find absent files
	for (const auto& p : dbHashFileList) {
		bool fileFound = false;
		for (const auto& q : newHashFileList) {
			if (eqfilenames(p.filename, q.filename))	{
				fileFound = true;
				break;
			}
		}
		if (!fileFound) {
			errors->push_back(std::make_unique<CE_AbsentFile>(p.filename));
		}
	}
}

void CompareErr::setDiffReason(DiffReason _reason)
{
	m_reason = _reason;
}

CompareErr::CompareErr(const std::string& _filename, DiffReason _reason)
{
	m_filename = QString::fromStdString(_filename/*.c_str()*/);
	setDiffReason(_reason);
}

QString CompareErr::text()
{
	switch (m_reason)
	{
	case DiffReason::HASH_DIFF:
		return QObject::tr("Checksums does not match");

	case DiffReason::NEW_FILE:
		return QObject::tr("New file");

	case DiffReason::ABSENT_FILE:
		return QObject::tr("File not found");

	default:
		return QObject::tr("Unknown");
	}
}

CE_HashDiff::CE_HashDiff(const std::string& _filename, const std::string& _newhash, const std::string& _dbhash) :
	CompareErr(_filename, DiffReason::HASH_DIFF)
{
	m_newhash = QString::fromStdString(_newhash);
	m_dbhash = QString::fromStdString(_dbhash);
}

QString CE_HashDiff::descrText()
{
	return tr("");
}

CE_NewFile::CE_NewFile(const std::string& _filename) :
	CompareErr(_filename, DiffReason::NEW_FILE)
{

}

QString CE_NewFile::descrText()
{
	return tr("");
}

CE_AbsentFile::CE_AbsentFile(const std::string& _filename) :
	CompareErr(_filename, DiffReason::ABSENT_FILE)
{

}

QString CE_AbsentFile::descrText()
{
	return tr("");
}

CE_IOError::CE_IOError(const std::string& _filename, std::string& _msg)	:
	CompareErr(_filename, DiffReason::IO_ERROR)
{
	m_msg = QDir::toNativeSeparators(QString::fromStdString(_msg));
}

QString CE_IOError::descrText()		
{
	return m_msg;
}
